from __future__ import unicode_literals

import logging

from mopidy import backend
from mopidy.models import Playlist

from . import Extension

logger = logging.getLogger(__name__)


class InternetArchivePlaylistsProvider(backend.PlaylistsProvider):

    def __init__(self, config, backend):
        super(InternetArchivePlaylistsProvider, self).__init__(backend)
        self._bookmarks = []
        self._username = config[Extension.ext_name]['username']
        self._playlist = Playlist(
            uri='%s://%s/bookmarks' % (Extension.ext_name, self._username),
            name='Internet Archive Bookmarks'
        )
        self.playlists = [self._playlist]

    def create(self, name):
        pass  # TODO

    def delete(self, uri):
        pass  # TODO

    def lookup(self, uri):
        for playlist in self.playlists:
            if playlist.uri == uri:
                return playlist
        return None

    def refresh(self):
        try:
            bookmarks = self.backend.client.bookmarks(self._username)
            logger.info('Loaded %d Internet Archive bookmarks', len(bookmarks))
            if bookmarks == self._bookmarks:
                return  # unchanged
            # FIXME: this can probably be removed when playlists return refs...
            tracks = self._tracks(bookmarks)
            logger.info('Loaded %d Internet Archive tracks', len(tracks))
            self.playlists = [self._playlist.copy(tracks=tracks)]
            backend.BackendListener.send('playlists_loaded')
            self._bookmarks = bookmarks
        except Exception as e:
            logger.error('Error loading Internet Archive bookmarks: %s', e)

    def save(self, playlist):
        pass  # TODO

    def _tracks(self, bookmarks):
        tracks = []
        for doc in bookmarks:
            tracks.extend(self.backend.library._lookup(doc['identifier']))
        return tracks



from __future__ import unicode_literals

from mopidy import backend


class InternetArchivePlaybackProvider(backend.PlaybackProvider):

    def change_track(self, track):
        track = track.copy(uri=self.backend.library.get_stream_url(track.uri))
        return super(InternetArchivePlaybackProvider, self).change_track(track)



from __future__ import unicode_literals

import collections
import requests

from urlparse import urljoin

BASE_URL = 'http://archive.org/'


class InternetArchiveClient(object):

    def __init__(self, base_url=BASE_URL, timeout=None):
        self._search_url = urljoin(base_url, '/advancedsearch.php')
        self._metadata_url = urljoin(base_url, '/metadata/')
        self._download_url = urljoin(base_url, '/download/')
        self._bookmarks_url = urljoin(base_url, '/bookmarks/')
        self._session = requests.Session()
        self._timeout = timeout

    def search(self, query, fields=None, sort=None, rows=None, start=None):
        response = self._session.get(self._search_url, params={
            'q': query,
            'fl[]': fields,
            'sort[]': sort,
            'rows': rows,
            'start': start,
            'output': 'json'
        }, timeout=self._timeout)
        if not response.content:
            raise self.SearchError(response.url)
        return self.SearchResult(response.json())

    def metadata(self, path):
        url = urljoin(self._metadata_url, path.lstrip('/'))
        response = self._session.get(url, timeout=self._timeout)
        data = response.json()

        if not data:
            raise LookupError('Internet Archive item %s not found' % path)
        elif 'error' in data:
            raise LookupError(data['error'])
        elif 'result' in data:
            return data['result']
        else:
            return data

    def bookmarks(self, username):
        url = urljoin(self._bookmarks_url, username + '?output=json')
        response = self._session.get(url, timeout=self._timeout)
        # requests for non-existant users yield text/xml response
        if response.headers['Content-Type'] != 'application/json':
            raise LookupError('Internet Archive user %s not found' % username)
        return response.json()

    def geturl(self, identifier, filename=None):
        if filename:
            return urljoin(self._download_url, identifier + '/' + filename)
        else:
            return urljoin(self._download_url, identifier + '/')

    class SearchResult(collections.Sequence):

        def __init__(self, result):
            self.query = result['responseHeader']['params']['q']
            self.rowcount = result['response']['numFound']
            self.docs = result['response']['docs']

        def __getitem__(self, key):
            return self.docs[key]

        def __len__(self):
            return len(self.docs)

        def __iter__(self):
            return iter(self.docs)

    class SearchError(Exception):

        pass



from __future__ import unicode_literals

import cachetools
import collections
import itertools
import logging
import operator
import re

from mopidy import backend
from mopidy.models import Album, Artist, Track, SearchResult, Ref
from uritools import uricompose, urisplit

from . import Extension
from .parsing import *  # noqa
from .query import Query

_URI_PREFIX = Extension.ext_name + ':'

_QUERY_CHAR_RE = re.compile(r'([+!(){}\[\]^"~*?:\\]|\&\&|\|\|)')

_QUERY_MAPPING = {
    'any': None,
    'album': 'title',
    'albumartist': 'creator',
    'date': 'date'
}

logger = logging.getLogger(__name__)


def _cache(cache_size=None, cache_ttl=None, **kwargs):
    """Cache factory"""
    if cache_size is None:
        return None  # mainly for testing/debugging
    elif cache_ttl is None:
        return cachetools.LRUCache(cache_size)
    else:
        return cachetools.TTLCache(cache_size, cache_ttl)


def _ref(metadata):
    identifier = metadata['identifier']
    uri = _URI_PREFIX + identifier
    name = metadata.get('title', identifier)
    if metadata.get('mediatype', 'collection') == 'collection':
        return Ref.directory(uri=uri, name=name)
    else:
        return Ref.album(uri=uri, name=name)


def _artists(metadata, default=[]):
    creator = metadata.get('creator')
    if not creator:
        return default
    elif isinstance(creator, basestring):
        return [Artist(name=creator)]
    else:
        return [Artist(name=name) for name in creator]


def _album(metadata, images=[]):
    identifier = metadata['identifier']
    uri = _URI_PREFIX + identifier
    name = metadata.get('title', identifier)
    artists = _artists(metadata)
    date = parse_date(metadata.get('date'))
    return Album(uri=uri, name=name, artists=artists, date=date, images=images)


def _track(metadata, file, album):
    identifier = metadata['identifier']
    filename = file['name']
    uri = _URI_PREFIX + identifier + '#' + filename
    name = file.get('title', filename)
    return Track(
        uri=uri,
        name=name,
        album=album,
        artists=_artists(file, album.artists),
        track_no=parse_track_no(file.get('track')),
        date=parse_date(file.get('date'), album.date),
        length=parse_length(file.get('length')),
        bitrate=parse_bitrate(file.get('bitrate')),
        # FIXME: comment=metadata.get('description'),
        last_modified=parse_mtime(file.get('mtime'))
    )


def _getfirstitem(object, keys, default=None):
    for key in keys:
        if key in object:
            return object[key]
    return default


def _quote(term):
    term = _QUERY_CHAR_RE.sub(r'\\\1', term)
    # only quote if term contains whitespace, since date:"2014-01-01"
    # will raise an error
    if any(c.isspace() for c in term):
        term = '"' + term + '"'
    return term


def _query(*args, **kwargs):
    terms = []
    for field, value in itertools.chain(args, kwargs.items()):
        if not value:
            continue
        elif isinstance(value, basestring):
            values = [_quote(value)]
        else:
            values = [_quote(v) for v in value]
        if len(values) == 1:
            term = values[0]
        else:
            term = '(%s)' % ' OR '.join(values)
        terms.append(field + ':' + term if field else term)
    return ' AND '.join(terms)


class InternetArchiveLibraryProvider(backend.LibraryProvider):

    root_directory = Ref.directory(
        uri=_URI_PREFIX,
        name='Internet Archive'
    )

    def __init__(self, config, backend):
        super(InternetArchiveLibraryProvider, self).__init__(backend)
        self._config = iaconfig = config[Extension.ext_name]
        self._search_filter = {
            'format': iaconfig['audio_formats'],
            '-collection': iaconfig['exclude_collections'],
            '-mediatype': iaconfig['exclude_mediatypes']
        }
        self._cache = _cache(**iaconfig)
        self._tracks = {}  # track cache for faster lookup

    def lookup(self, uri):
        try:
            return [self._tracks[uri]]
        except KeyError:
            logger.debug("track lookup cache miss %r", uri)
        try:
            _, _, identifier, _, filename = urisplit(uri)
            tracks = self._lookup(identifier)
            self._tracks = trackmap = {t.uri: t for t in tracks}
            return [trackmap[uri]] if filename else tracks
        except Exception as e:
            logger.error('Lookup failed for %s: %s', uri, e)
            return []

    def browse(self, uri):
        try:
            identifier = urisplit(uri).path
            if not identifier:
                return self._browse_root()
            elif identifier in self._config['collections']:
                return self._browse_collection(identifier)
            else:
                return self._browse_item(identifier)
        except Exception as e:
            logger.error('Error browsing %s: %s', uri, e)
            return []

    def search(self, query=None, uris=None):
        try:
            terms = []
            for field, value in (query.iteritems() if query else []):
                if field not in _QUERY_MAPPING:
                    return None  # no result if unmapped field
                elif isinstance(value, basestring):
                    terms.append((_QUERY_MAPPING[field], value))
                else:
                    terms.extend((_QUERY_MAPPING[field], v) for v in value)
            if uris:
                ids = filter(None, (urisplit(uri).path for uri in uris))
                terms.append(('collection', tuple(ids)))
            return self._search(*terms)
        except Exception as e:
            logger.error('Error searching the Internet Archive: %s', e)
            return None

    def find_exact(self, query=None, uris=None):
        try:
            terms = []
            for field, value in (query.iteritems() if query else []):
                if field not in _QUERY_MAPPING:
                    return None  # no result if unmapped field
                elif isinstance(value, basestring):
                    terms.append((_QUERY_MAPPING[field], value))
                else:
                    terms.extend((_QUERY_MAPPING[field], v) for v in value)
            if uris:
                ids = filter(None, (urisplit(uri).path for uri in uris))
                terms.append(('collection', tuple(ids)))
            result = self._search(*terms)
            albums = filter(Query(query, True).match_album, result.albums)
            return result.copy(albums=albums)
        except Exception as e:
            logger.error('Error searching the Internet Archive: %s', e)
            return None

    def refresh(self, uri=None):
        logger.info('Clearing Internet Archive cache')
        self._cache.clear()
        self._tracks.clear()

    def get_stream_url(self, uri):
        _, _, identifier, _, filename = urisplit(uri)
        return self.backend.client.geturl(identifier, filename)

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_root(self):
        collections = self._config['collections']
        result = self.backend.client.search(
            _query(identifier=collections, mediatype='collection'),
            fields=['identifier', 'title', 'mediatype'],
            rows=len(collections)
        )
        # return in same order as listed in config
        docs = {doc['identifier']: doc for doc in result}
        refs = []
        for id in collections:
            if id in docs:
                refs.append(_ref(docs[id]))
            else:
                logger.warn('Internet Archive collection "%s" not found', id)
        return refs

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_collection(self, identifier):
        result = self.backend.client.search(
            _query(collection=identifier, **self._search_filter),
            fields=['identifier', 'title', 'mediatype'],
            sort=self._config['browse_order'],
            rows=self._config['browse_limit']
        )
        return map(_ref, result)

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _browse_item(self, identifier):
        tracks = self._lookup(identifier)
        self._tracks = {t.uri: t for t in tracks}  # cache tracks
        return [Ref.track(uri=t.uri, name=t.name) for t in tracks]

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _lookup(self, identifier):
        client = self.backend.client
        item = client.metadata(identifier)
        # TODO: original vs. derived, "tmp" values?
        files = collections.defaultdict(list)
        for file in item['files']:
            files[file['format']].append(file)
        images = []
        for file in _getfirstitem(files, self._config['image_formats'], []):
            images.append(client.geturl(identifier, file['name']))
        album = _album(item['metadata'], images)
        tracks = []
        for file in _getfirstitem(files, self._config['audio_formats'], []):
            tracks.append(_track(item['metadata'], file, album))
        # sort tracks by track_no if given, by uri/filename otherwise
        tracks.sort(key=lambda t: (t.track_no or 0, t.uri))
        return tracks

    @cachetools.cachedmethod(operator.attrgetter('_cache'))
    def _search(self, *terms):
        result = self.backend.client.search(
            _query(*terms, **self._search_filter),
            fields=['identifier', 'title', 'creator', 'date'],
            sort=self._config['search_order'],
            rows=self._config['search_limit']
        )
        uri = uricompose(Extension.ext_name, query=result.query)
        return SearchResult(uri=uri, albums=[_album(doc) for doc in result])



from __future__ import unicode_literals

import unittest

from mopidy.models import Artist, Album, Track, Ref
from mopidy_podcast.query import Query


def anycase(*strings):
    return [s.upper() for s in strings] + [s.lower() for s in strings]


class QueryTest(unittest.TestCase):

    def assertQueryMatches(self, model, **kwargs):
        query = Query(kwargs, exact=False)
        self.assertTrue(query.match(model))

    def assertNotQueryMatches(self, model, **kwargs):
        query = Query(kwargs, exact=False)
        self.assertFalse(query.match(model))

    def assertQueryMatchesExact(self, model, **kwargs):
        query = Query(kwargs, exact=True)
        self.assertTrue(query.match(model))

    def assertNotQueryMatchesExact(self, model, **kwargs):
        query = Query(kwargs, exact=True)
        self.assertFalse(query.match(model))

    def test_create_query(self):
        for exact in (True, False):
            q1 = Query(dict(any='foo'), exact)
            self.assertEqual(q1.exact, exact)
            self.assertEqual(len(q1), 1)
            self.assertItemsEqual(q1, ['any'])
            self.assertEqual(len(q1['any']), 1)
            self.assertEqual(q1['any'][0], 'foo')
            self.assertNotEqual(q1['any'][0], 'bar')

            q2 = Query(dict(any=['foo', 'bar'], artist='x'), exact)
            self.assertEqual(q2.exact, exact)
            self.assertEqual(len(q2), 2)
            self.assertItemsEqual(q2, ['any', 'artist'])
            self.assertEqual(len(q2['any']), 2)
            self.assertEqual(len(q2['artist']), 1)
            self.assertEqual(q2['any'][0], 'foo')
            self.assertEqual(q2['any'][1], 'bar')
            self.assertEqual(q2['artist'][0], 'x')

    def test_query_errors(self):
        for exact in (True, False):
            with self.assertRaises(LookupError):
                Query(None, exact)
            with self.assertRaises(LookupError):
                Query({}, exact)
            with self.assertRaises(LookupError):
                Query({'artist': None}, exact)
            with self.assertRaises(LookupError):
                Query({'artist': ''}, exact)
            with self.assertRaises(LookupError):
                Query({'artist': []}, exact)
            with self.assertRaises(LookupError):
                Query({'artist': ['']}, exact)
            with self.assertRaises(LookupError):
                Query({'any': None}, exact)
            with self.assertRaises(LookupError):
                Query({'any': ''}, exact)
            with self.assertRaises(LookupError):
                Query({'any': []}, exact)
            with self.assertRaises(LookupError):
                Query({'any': ['']}, exact)
            with self.assertRaises(LookupError):
                Query({'foo': 'bar'}, exact)
            with self.assertRaises(TypeError):
                q = Query(dict(any='foo'), exact)
                q.match(Ref(name='foo'))

    def test_match_artist(self):
        artist = Artist(name='foo')

        for name in anycase('f', 'o', 'fo', 'oo', 'foo'):
            self.assertQueryMatches(artist, any=name)
            self.assertQueryMatches(artist, artist=name)

        self.assertQueryMatchesExact(artist, any='foo')
        self.assertQueryMatchesExact(artist, artist='foo')

        self.assertNotQueryMatches(artist, any='none')
        self.assertNotQueryMatches(artist, artist='none')

        self.assertNotQueryMatchesExact(artist, any='none')
        self.assertNotQueryMatchesExact(artist, artist='none')

    def test_match_album(self):
        artist = Artist(name='foo')
        album = Album(name='bar', artists=[artist])

        for name in anycase('f', 'o', 'fo', 'oo', 'foo'):
            self.assertQueryMatches(album, any=name)
            self.assertQueryMatches(album, artist=name)
            self.assertQueryMatches(album, albumartist=name)
        for name in anycase('b', 'a', 'ba', 'ar', 'bar'):
            self.assertQueryMatches(album, any=name)
            self.assertQueryMatches(album, album=name)

        self.assertQueryMatchesExact(album, any='foo')
        self.assertQueryMatchesExact(album, artist='foo')
        self.assertQueryMatchesExact(album, albumartist='foo')
        self.assertQueryMatchesExact(album, any='bar')
        self.assertQueryMatchesExact(album, album='bar')

        self.assertNotQueryMatches(album, any='none')
        self.assertNotQueryMatches(album, artist='bar')
        self.assertNotQueryMatches(album, album='foo')

        self.assertNotQueryMatchesExact(album, any='none')
        self.assertNotQueryMatchesExact(album, artist='bar')
        self.assertNotQueryMatchesExact(album, album='foo')

    def test_match_track(self):
        artist = Artist(name='foo')
        album = Album(name='bar', artists=[Artist(name='v/a')])
        track = Track(name='zyx', album=album, artists=[artist])

        for name in anycase('f', 'o', 'fo', 'oo', 'foo'):
            self.assertQueryMatches(track, any=name)
            self.assertQueryMatches(track, artist=name)
        for name in anycase('b', 'a', 'ba', 'ar', 'bar'):
            self.assertQueryMatches(track, any=name)
            self.assertQueryMatches(track, album=name)
        for name in anycase('v', '/', 'v/', '/a', 'v/a'):
            self.assertQueryMatches(track, any=name)
            self.assertQueryMatches(track, albumartist=name)
        for name in anycase('z', 'y', 'zy', 'yx', 'zyx'):
            self.assertQueryMatches(track, any=name)
            self.assertQueryMatches(track, track_name=name)

        self.assertQueryMatchesExact(track, any='foo')
        self.assertQueryMatchesExact(track, artist='foo')
        self.assertQueryMatchesExact(track, any='bar')
        self.assertQueryMatchesExact(track, album='bar')
        self.assertQueryMatchesExact(track, any='v/a')
        self.assertQueryMatchesExact(track, albumartist='v/a')
        self.assertQueryMatchesExact(track, any='zyx')
        self.assertQueryMatchesExact(track, track_name='zyx')

        self.assertNotQueryMatches(track, any='none')
        self.assertNotQueryMatches(track, artist='bar')
        self.assertNotQueryMatches(track, album='foo')
        self.assertNotQueryMatches(track, albumartist='zyx')
        self.assertNotQueryMatches(track, track_name='v/a')

        self.assertNotQueryMatchesExact(track, any='none')
        self.assertNotQueryMatchesExact(track, artist='bar')
        self.assertNotQueryMatchesExact(track, album='foo')
        self.assertNotQueryMatchesExact(track, albumartist='zyx')
        self.assertNotQueryMatchesExact(track, track_name='v/a')



from __future__ import unicode_literals

import os

from mopidy import config, ext

__version__ = '1.0.2'

SORT_FIELDS = ['%s %s' % (f, o) for o in ('asc', 'desc') for f in (
    'addeddate',
    'avg_rating',
    'call_number',
    'createdate',
    'creatorSorter',
    'date',
    'downloads',
    'foldoutcount',
    'headerImage',
    'identifier',
    'imagecount',
    'indexdate',
    'languageSorter',
    'licenseurl',
    'month',
    'nav_order',
    'num_reviews',
    'publicdate',
    'reviewdate',
    'stars',
    'titleSorter',
    'week',
    'year'
)]


class Extension(ext.Extension):

    dist_name = 'Mopidy-InternetArchive'
    ext_name = 'internetarchive'
    version = __version__

    def get_default_config(self):
        return config.read(os.path.join(os.path.dirname(__file__), 'ext.conf'))

    def get_config_schema(self):
        schema = super(Extension, self).get_config_schema()
        schema['base_url'] = config.String()
        schema['username'] = config.String(optional=True)
        schema['collections'] = config.List()
        schema['audio_formats'] = config.List()
        schema['image_formats'] = config.List()
        schema['browse_limit'] = config.Integer(minimum=1, optional=True)
        schema['browse_order'] = config.String(choices=SORT_FIELDS, optional=True)  # noqa
        schema['search_limit'] = config.Integer(minimum=1, optional=True)
        schema['search_order'] = config.String(choices=SORT_FIELDS, optional=True)  # noqa
        schema['exclude_collections'] = config.List(optional=True)
        schema['exclude_mediatypes'] = config.List(optional=True)
        schema['cache_size'] = config.Integer(minimum=1, optional=True)
        schema['cache_ttl'] = config.Integer(minimum=0, optional=True)
        schema['timeout'] = config.Integer(minimum=0, optional=True)

        # no longer used (for now)
        schema['bookmarks'] = config.Deprecated()
        schema['bookmarks_label'] = config.Deprecated()
        schema['browse_label'] = config.Deprecated()
        schema['excludes'] = config.Deprecated()
        schema['mediatypes'] = config.Deprecated()
        schema['sort_order'] = config.Deprecated()

        return schema

    def setup(self, registry):
        from .backend import InternetArchiveBackend
        registry.add('backend', InternetArchiveBackend)



from __future__ import unicode_literals

import collections

from mopidy.models import Artist, Album, Track

QUERY_FIELDS = {
    'uri',
    'track_name',
    'album',
    'artist',
    'composer',
    'performer',
    'albumartist',
    'genre',
    'date',
    'comment',
    'track_no',
    'any'
}

DEFAULT_FILTERS = dict.fromkeys(QUERY_FIELDS, lambda q, v: False)

TRACK_FILTERS = [
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, track: bool(
            track.uri and q in track.uri.lower()
        ),
        track_name=lambda q, track: bool(
            track.name and q in track.name.lower()
        ),
        album=lambda q, track: bool(
            track.album and track.album.name and q in track.album.name.lower()
        ),
        artist=lambda q, track: any(
            a.name and q in a.name.lower() for a in track.artists
        ),
        composer=lambda q, track: any(
            a.name and q in a.name.lower() for a in track.composers
        ),
        performer=lambda q, track: any(
            a.name and q in a.name.lower() for a in track.performers
        ),
        albumartist=lambda q, track: track.album and any(
            a.name and q in a.name.lower() for a in track.album.artists
        ),
        genre=lambda q, track: bool(
            track.genre and q in track.genre.lower()
        ),
        date=lambda q, track: bool(
            track.date and track.date.startswith(q)
        ),
        comment=lambda q, track: bool(
            track.comment and q in track.comment.lower(),
        ),
        track_no=lambda q, track: q.isdigit() and int(q) == track.track_no
    ),
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, track: q == track.uri,
        track_name=lambda q, track: q == track.name,
        album=lambda q, track: track.album and q == track.album.name,
        artist=lambda q, track: any(
            q == a.name for a in track.artists
        ),
        composer=lambda q, track: any(
            q == a.name for a in track.composers
        ),
        performer=lambda q, track: any(
            q == a.name for a in track.performers
        ),
        albumartist=lambda q, track: track.album and any(
            q == a.name for a in track.album.artists
        ),
        genre=lambda q, track: q == track.genre,
        date=lambda q, track: q == track.date,
        comment=lambda q, track: q == track.comment,
        track_no=lambda q, track: q.isdigit() and int(q) == track.track_no
    )
]

ALBUM_FILTERS = [
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, album: bool(album.uri and q in album.uri.lower()),
        album=lambda q, album: bool(album.name and q in album.name.lower()),
        artist=lambda q, album: any(
            a.name and q in a.name.lower() for a in album.artists
        ),
        albumartist=lambda q, album: any(
            a.name and q in a.name.lower() for a in album.artists
        ),
        date=lambda q, album: bool(
            album.date and album.date.startswith(q)
        )
    ),
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, album: q == album.uri,
        album=lambda q, album: q == album.name,
        artist=lambda q, album: any(
            q == a.name for a in album.artists
        ),
        albumartist=lambda q, album: any(
            q == a.name for a in album.artists
        ),
        date=lambda q, album: q == album.date
    )
]

ARTIST_FILTERS = [
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, artist: bool(artist.uri and q in artist.uri.lower()),
        artist=lambda q, artist: bool(artist.name and q in artist.name.lower())
    ),
    dict(
        DEFAULT_FILTERS,
        uri=lambda q, artist: q == artist.uri,
        artist=lambda q, artist: q == artist.name
    )
]


class Query(collections.Mapping):

    _track_filter = None
    _album_filter = None
    _artist_filter = None

    def __init__(self, terms, exact=False):
        if not terms:
            raise LookupError('Empty query not allowed')
        self.terms = {}
        self.exact = exact

        for field, values in terms.iteritems():
            if field not in QUERY_FIELDS:
                raise LookupError('Invalid query field %r' % field)
            if isinstance(values, basestring):
                values = [values]
            if not (values and all(values)):
                raise LookupError('Missing query value for %r' % field)
            if exact:
                self.terms[field] = values
            else:
                self.terms[field] = [v.lower() for v in values]

    def __getitem__(self, key):
        return self.terms[key]

    def __iter__(self):
        return iter(self.terms)

    def __len__(self):
        return len(self.terms)

    def match(self, model):
        if isinstance(model, Track):
            return self.match_track(model)
        elif isinstance(model, Album):
            return self.match_album(model)
        elif isinstance(model, Artist):
            return self.match_artist(model)
        else:
            raise TypeError('Invalid model type: %s' % type(model))

    def match_artist(self, artist):
        if not self._artist_filter:
            self._artist_filter = self._filter(ARTIST_FILTERS[self.exact])
        return self._artist_filter(artist)

    def match_album(self, album):
        if not self._album_filter:
            self._album_filter = self._filter(ALBUM_FILTERS[self.exact])
        return self._album_filter(album)

    def match_track(self, track):
        if not self._track_filter:
            self._track_filter = self._filter(TRACK_FILTERS[self.exact])
        return self._track_filter(track)

    def _filter(self, filtermap):
        from functools import partial
        filters = []
        for field, values in self.terms.iteritems():
            filters.extend(partial(filtermap[field], v) for v in values)

        def func(model):
            return all(f(model) for f in filters)
        return func


# setup 'any' filters
def _any_filter(filtermap):
    filters = filtermap.values()

    def any_filter(q, v):
        return any(f(q, v) for f in filters)
    return any_filter

for i in (False, True):
    TRACK_FILTERS[i]['any'] = _any_filter(TRACK_FILTERS[i])
    ALBUM_FILTERS[i]['any'] = _any_filter(ALBUM_FILTERS[i])
    ARTIST_FILTERS[i]['any'] = _any_filter(ARTIST_FILTERS[i])



from __future__ import unicode_literals

import unittest

from mopidy_internetarchive.parsing import *  # noqa


class ParsingTest(unittest.TestCase):

    def test_parse_bitrate(self):
        self.assertEqual(parse_bitrate(None), None)
        self.assertEqual(parse_bitrate(''), None)
        self.assertEqual(parse_bitrate([]), None)
        self.assertEqual(parse_bitrate('tmp'), None)

        self.assertEqual(parse_bitrate(None, 42), 42)
        self.assertEqual(parse_bitrate('', 42), 42)
        self.assertEqual(parse_bitrate([], 42), 42)
        self.assertEqual(parse_bitrate('tmp', 42), 42)

        self.assertEqual(parse_bitrate('0'), 0)
        self.assertEqual(parse_bitrate('1'), 1)
        self.assertEqual(parse_bitrate('42.123'), 42)

    def test_parse_date(self):
        self.assertEqual(parse_date(None), None)
        self.assertEqual(parse_date(''), None)
        self.assertEqual(parse_date([]), None)
        self.assertEqual(parse_date('tmp'), None)

        self.assertEqual(parse_date(None, 42), 42)
        self.assertEqual(parse_date('', 42), 42)
        self.assertEqual(parse_date([], 42), 42)
        self.assertEqual(parse_date('tmp', 42), 42)

        self.assertEqual(parse_date('2014-02-21T06:00:00Z'), '2014-02-21')
        self.assertEqual(parse_date('2014-02-21'), '2014-02-21')
        self.assertEqual(parse_date('2014-02'), '2014-02-01')
        self.assertEqual(parse_date('2014'), '2014-01-01')

    def test_parse_length(self):
        self.assertEqual(parse_length(None), None)
        self.assertEqual(parse_length(''), None)
        self.assertEqual(parse_length([]), None)
        self.assertEqual(parse_length('tmp'), None)

        self.assertEqual(parse_length(None, 42), 42)
        self.assertEqual(parse_length('', 42), 42)
        self.assertEqual(parse_length([], 42), 42)
        self.assertEqual(parse_length('tmp', 42), 42)

        self.assertEqual(parse_length('0'), 0)
        self.assertEqual(parse_length('1'), 1 * 1000)
        self.assertEqual(parse_length('60'), 60 * 1000)
        self.assertEqual(parse_length('1:00'), 60 * 1000)
        self.assertEqual(parse_length('60:00'), 3600 * 1000)
        self.assertEqual(parse_length('1:00:00'), 3600 * 1000)

    def test_parse_mtime(self):
        self.assertEqual(parse_mtime(None), None)
        self.assertEqual(parse_mtime(''), None)
        self.assertEqual(parse_mtime([]), None)
        self.assertEqual(parse_mtime('tmp'), None)

        self.assertEqual(parse_mtime(None, 42), 42)
        self.assertEqual(parse_mtime('', 42), 42)
        self.assertEqual(parse_mtime([], 42), 42)
        self.assertEqual(parse_mtime('tmp', 42), 42)

        self.assertEqual(parse_mtime('0'), 0)
        self.assertEqual(parse_mtime('1'), 1)

    def test_parse_track_no(self):
        self.assertEqual(parse_track_no(None), None)
        self.assertEqual(parse_track_no(''), None)
        self.assertEqual(parse_track_no([]), None)
        self.assertEqual(parse_track_no('tmp'), None)

        self.assertEqual(parse_track_no(None, 42), 42)
        self.assertEqual(parse_track_no('', 42), 42)
        self.assertEqual(parse_track_no([], 42), 42)
        self.assertEqual(parse_track_no('tmp', 42), 42)

        self.assertEqual(parse_track_no('0'), 0)
        self.assertEqual(parse_track_no('1'), 1)
        self.assertEqual(parse_track_no('1/100'), 1)
        self.assertEqual(parse_track_no('1/tmp'), 1)
